# -*- coding: utf-8 -*-

#  Copyright (c) 2021 Dafne-Imaging Team
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import math
import numpy as np

# new implementation, working on a generic number of axes
def padorcut(arrayin, newSize, axis = None):
    nDims = arrayin.ndim
    
    # extend dimensions
    while nDims < len(newSize):
        arrayin = np.expand_dims(arrayin, nDims)
        nDims = arrayin.ndim
        
    if type(axis) is int:
        # check if newsz is iterable, otherwise assume it's a number
        try:
            newSz = newSize[axis]
        except:
            newSz = int(newSize)
        oldSz = arrayin.shape[axis]
        if oldSz < newSz:
            padBefore = int(math.floor(float(newSz - oldSz)/2))
            padAfter = int(math.ceil(float(newSz - oldSz)/2))
            padding = []
            for i in range(nDims):
                if i == axis:
                    padding.append( (padBefore, padAfter) )
                else:
                    padding.append( (0,0) )
            return np.pad(arrayin, padding, 'constant')
        elif oldSz > newSz:
            cutBefore = int(math.floor(float(oldSz - newSz)/2))
            cutAfter = int(math.ceil(float(oldSz - newSz)/2))
            slc = [slice(None)]*nDims
            slc[axis] = slice(cutBefore, -cutAfter)
            return arrayin[tuple(slc)]
        else:
            return arrayin
    else:
        for ax in range(nDims):
            arrayin = padorcut(arrayin, newSize, ax)
        return arrayin

def translate(arrayin, translation):
    padBeforeX = translation[0] if translation[0] > 0 else 0
    padAfterX = -translation[0] if translation[0] < 0 else 0
    padBeforeY = translation[1] if translation[1] > 0 else 0
    padAfterY = -translation[1] if translation[1] < 0 else 0
    
    cutBeforeX = padAfterX
    cutAfterX = padBeforeX
    cutBeforeY = padAfterY
    cutAfterY = padBeforeY
    
    nx, ny = arrayin.shape    
    
    cutArray = arrayin[cutBeforeX:(nx-cutAfterX), cutBeforeY:(ny-cutAfterY)]
    arrayout = arrayout=np.pad(cutArray, ( (padBeforeX,padAfterX), (padBeforeY, padAfterY) ), 'constant')
    return arrayout
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2021 Dafne-Imaging Team
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import time

import matplotlib
import matplotlib.pyplot as plt
import voxel

import pydicom as dicom

import numpy as np
import os
import sys
import traceback

try:
    CursorShape = matplotlib.backends.backend_qt.QtCore.Qt.CursorShape
except:
    #print("CursorShape not found")
    class CursorShape:
        ArrowCursor = 0

try:
    from matplotlib.backend_bases import MouseButton
except:
    #print("Mousebutton not found")
    class MouseButton:
        LEFT = 1
        RIGHT = 3

from ..misc import dosma_volume_from_path

DEFAULT_INTERPOLATION = 'spline36'
# DEFAULT_INTERPOLATION = 'none' # DEBUG
INVERT_SCROLL = True
DO_DEBUG = False
MOUSE_SCROLL_DEBOUNCE_TIME = 0.1


class ImListProxy:

    def __init__(self, medical_volume):
        self.medical_volume = medical_volume

    def __getitem__(self, item):
        return self.medical_volume.volume[:, :, item].astype(np.float32)

    def __len__(self):
        return self.medical_volume.shape[2]


class ImageShow:
    contrastWindow = None
    channelBalance = np.array([1.0, 1.0, 1.0])

    def __init__(self, im=None, axes=None, window=None, cmap=None, debounce_time=MOUSE_SCROLL_DEBOUNCE_TIME):
        ImageShow.contrastWindow = window

        self.imPlot = None

        if axes is None:
            # initialize the figure
            self.fig = plt.figure()
            self.axes = self.fig.add_subplot(111)
        else:
            self.axes = axes
            self.fig = self.axes.get_figure()

        self.axes.axis('off')
        self.connect_ids = []
        self.connectSignals()

        # stack of images
        self.imList = []
        self.dicomHeaderList = None
        self.curImage = None
        self.cmap = cmap
        self.isImageRGB = False
        self.basepath = ''
        self.basename = ''

        self.resolution = [1, 1, 1]
        self.resolution_valid = False
        self.affine = None
        self.medical_volume = None
        self.scroll_debounce_time = debounce_time
        self.last_scroll_time = 0

        self.interpolation = DEFAULT_INTERPOLATION

        # These methods can be defined in a subclass and called when some event occurs
        # self.leftPressCB = None
        # self.leftMoveCB = None
        # self.leftReleaseCB = None
        # self.refreshCB = None

        self.oldMouseXY = (0, 0)
        self.startXY = None

        if im is not None:
            if type(im) is np.ndarray:
                self.loadNumpyArray(im)
            elif isinstance(im, voxel.MedicalVolume):
                self.loadNumpyArray(im.volume)
            elif type(im) is str:
                if os.path.isdir(im):
                    self.loadDirectory(im)
                else:
                    try:
                        im = self.loadDicomFile(im)
                    except:
                        pass
                    self.imList.append(im)
            self.curImage = 0
            self.displayImage(0)

    def connectSignals(self):
        if self.connect_ids: return
        self.connect_ids = []
        self.connect_ids.append(self.fig.canvas.mpl_connect('button_press_event', self.btnPressCB))
        self.connect_ids.append(self.fig.canvas.mpl_connect('button_release_event', self.btnReleaseCB))
        self.connect_ids.append(self.fig.canvas.mpl_connect('motion_notify_event', self.mouseMoveCB))
        self.connect_ids.append(self.fig.canvas.mpl_connect('scroll_event', self.mouseScrollCB))
        self.connect_ids.append(self.fig.canvas.mpl_connect('key_press_event', self.keyPressCB))
        self.connect_ids.append(self.fig.canvas.mpl_connect('key_release_event', self.keyReleaseCB))

    def disconnectSignals(self):
        for cid in self.connect_ids:
            self.fig.canvas.mpl_disconnect(cid)
        self.connect_ids = []

    def displayImageRGB(self):
        # print "Displaying image"
        dispImage = np.copy(self.image)
        dispImage[:, :, 0] *= ImageShow.channelBalance[0]
        dispImage[:, :, 1] *= ImageShow.channelBalance[1]
        dispImage[:, :, 2] *= ImageShow.channelBalance[2]
        dispImage = (dispImage - ImageShow.contrastWindow[0]) / (
                    ImageShow.contrastWindow[1] - ImageShow.contrastWindow[0])
        dispImage[dispImage < 0] = 0
        dispImage[dispImage > 1] = 1
        if self.imPlot is None:
            self.imPlot = self.axes.imshow(dispImage, interpolation=self.interpolation,
                                           aspect=self.resolution[0] / self.resolution[1])
        else:
            self.imPlot.set_data(dispImage)
        self.redraw()

    def displayImage(self, im, cmap=None, redraw=True):
        if cmap is None:
            if self.cmap is None:
                cmap = 'gray'
            else:
                cmap = self.cmap

        try:
            oldSize = self.image.shape
        except:
            oldSize = (-1, -1)

        # im can be an integer index in the imList
        if im is None:
            try:
                self.imPlot.set_data(np.zeros((0, 0)))
            except:
                pass
            try:
                self.axes.set_title('')
            except:
                pass
            return
        if isinstance(im, int):
            if im >= 0 and im < len(self.imList):
                self.curImage = im
                self.image = self.imList[im]
        else:  # otherwise let's assume it is pixel data
            self.image = im

        title = ''
        try:
            title = self.instructions + '\n'
        except:
            pass

        title += 'Image: %d' % self.curImage

        try:
            self.axes.set_title(title)
        except:
            pass

        # calculate the contrast if it was not already defined
        if ImageShow.contrastWindow is None:
            ImageShow.contrastWindow = self.calcContrast(self.image)

        if self.image.ndim == 3:
            self.isImageRGB = True
            self.displayImageRGB()
            return
        else:
            self.isImageRGB = False

        self.setCmap(cmap, False)

        if self.imPlot:
            if oldSize != self.image.shape:  # if the image shape is different, force a new imPlot to be created
                try:
                    self.imPlot.remove()
                except:
                    pass
                self.imPlot = None

        # Create the image plot if there is none; otherwise update the data in the existing frame (faster)
        if self.imPlot is None:
            self.imPlot = self.axes.imshow(self.image, interpolation=self.interpolation,
                                           vmin=ImageShow.contrastWindow[0],
                                           vmax=ImageShow.contrastWindow[1],
                                           cmap=self.cmap, zorder=-1, aspect=self.resolution[0] / self.resolution[1])
        else:
            self.imPlot.set_data(self.image)

        if redraw:
            self.redraw()

    def refreshCB(self):
        pass

    def redraw(self):
        try:
            self.refreshCB()
        except Exception as err:
            if DO_DEBUG: traceback.print_exc()
        self.fig.canvas.draw()

    def mouseScrollCB(self, event):
        current_time = time.time()
        if current_time - self.last_scroll_time < self.scroll_debounce_time:
            return
        self.last_scroll_time = current_time
        self.oldMouseXY = (event.x, event.y)  # this will suppress the mouse move event
        step = np.sign(event.step) # new versions of matplotlib use steps of eight of degree and can generate very large numbers. Mac is also weird
        
        step = -step if INVERT_SCROLL else step
        if self.curImage is None or (step > 0 and self.curImage == 0) or (
                step < 0 and self.curImage > len(self.imList) - 1):
            return

        if event.inaxes != self.axes:
            return

        self.curImage = self.curImage - step
        if self.curImage < 0:
            self.curImage = 0
        if self.curImage > len(self.imList) - 1:
            self.curImage = len(self.imList) - 1
        self.displayImage(self.imList[int(self.curImage)], self.cmap, redraw=False)
        self.redraw()
        try:
            self.fig.canvas.setFocus()
        except:
            pass

    def keyReleaseCB(self, event):
        pass

    def keyPressCB(self, event):
        event.step = 0
        if event.key == 'right' or event.key == 'down':
            event.step = 1 if INVERT_SCROLL else -1
        elif event.key == 'left' or event.key == 'up':
            event.step = -1 if INVERT_SCROLL else 1
        if event.step != 0:
            self.mouseScrollCB(event)

    def isCursorNormal(self):
        try:
            cursor = self.fig.canvas.cursor().shape()
            # cursor. 0 is the arrow, which means we are not zooming or panning.
        except:
            return True

        if isinstance(CursorShape.ArrowCursor, int) and not isinstance(cursor, int):
            return cursor.value == CursorShape.ArrowCursor
        else:
            return cursor == CursorShape.ArrowCursor

    def leftPressCB(self, event):
        pass

    def btnPressCB(self, event):
        if not self.isCursorNormal():
            return
        if event.button is MouseButton.LEFT:
            try:
                self.leftPressCB(event)
            except Exception as err:
                if DO_DEBUG: traceback.print_exc()
        if event.button is MouseButton.RIGHT:
            if event.dblclick:
                self.resetContrast()
            else:
                self.startContrast = ImageShow.contrastWindow
                self.startBalance = np.copy(ImageShow.channelBalance)
                self.startXY = (event.x, event.y)
                self.rightPressCB(event)

    def rightPressCB(self, event):
        pass

    def leftReleaseCB(self, event):
        pass

    def btnReleaseCB(self, event):
        if event.button is MouseButton.LEFT:
            try:
                self.leftReleaseCB(event)
            except Exception as err:
                if DO_DEBUG: traceback.print_exc()
        if event.button is MouseButton.RIGHT:
            self.imPlot.set_interpolation(self.interpolation)
            self.startXY = None  #
            self.redraw()
            self.rightReleaseCB(event)

    def rightReleaseCB(self, event):
        pass

    def resetContrast(self):
        print("Resetting contrast")
        ImageShow.contrastWindow = self.calcContrast(self.image)
        if not self.isImageRGB:
            self.imPlot.set_clim(ImageShow.contrastWindow)
            self.redraw()
        else:
            # if image is RGB, we need to redraw it completely. Maybe it will be too slow?
            self.displayImageRGB()

    def leftMoveCB(self, event):
        pass

    # callback for mouse move
    def mouseMoveCB(self, event):
        xy = (event.x, event.y)
        if xy == self.oldMouseXY: return  # reject mouse move events when the mouse doesn't move
        self.oldMouseXY = xy
        if event.button is MouseButton.LEFT:
            try:
                self.leftMoveCB(event)
            except Exception as err:
                if DO_DEBUG: traceback.print_exc()
        if event.button is not MouseButton.RIGHT or self.startXY is None:
            return

        self.imPlot.set_interpolation('none')
        # convert contrast limits into window center and size
        contrastCenter = (self.startContrast[0] + self.startContrast[1]) / 2
        contrastExtent = (self.startContrast[1] - self.startContrast[0]) / 2

        # calculate displacemente of the mouse
        xDisplacement = event.x - self.startXY[0]
        yDisplacement = event.y - self.startXY[1]

        if event.key == 'control':
            ImageShow.channelBalance[0] = self.startBalance[0] - (
                        float(xDisplacement) / 100 + float(yDisplacement) / 100)
            ImageShow.channelBalance[1] = self.startBalance[1] + float(xDisplacement) / 100
            ImageShow.channelBalance[2] = self.startBalance[2] + float(yDisplacement) / 100
            ImageShow.channelBalance[ImageShow.channelBalance < 0] = 0
            ImageShow.channelBalance[ImageShow.channelBalance > 1] = 1
            # print ImageShow.channelBalance
        else:
            # recalculate the window
            # the displacements have negative sign because it feels more natural
            contrastCenter = contrastCenter - yDisplacement
            contrastExtent = contrastExtent - xDisplacement
            if contrastExtent < 1:
                contrastExtent = 1

            # set the contrast window
            ImageShow.contrastWindow = (contrastCenter - contrastExtent, contrastCenter + contrastExtent)

        if not self.isImageRGB:
            self.imPlot.set_clim(ImageShow.contrastWindow)
            self.redraw()
        else:
            # if image is RGB, we need to redraw it completely. Maybe it will be too slow?
            self.displayImageRGB()

    def calcContrast(self, im):
        maxVal = np.percentile(im, 90)
        if maxVal <= 1: maxVal = np.max(im.flat)
        if maxVal == 0: maxVal = 1
        return (0, maxVal)  # stretch contrast to remove outliers

    def setCmap(self, cmap, redraw=True):
        self.cmap = cmap;
        if self.imPlot is not None:
            self.imPlot.set_cmap(cmap)

        if redraw:
            self.redraw()

    def getDicomResolution(self, ds):
        resolution_valid = False
        try:
            slThickness = ds.SpacingBetweenSlices
        except:
            try:
                slThickness = ds.SliceThickness
            except:
                slThickness = 1

        try:
            pixelSpacing = ds.PixelSpacing
            resolution_valid = True
        except:
            pixelSpacing = [1, 1]

        resolution = [float(pixelSpacing[0]), float(pixelSpacing[1]), float(slThickness)]
        return resolution, resolution_valid

    def loadDicomFile(self, fname):
        ds = dicom.dcmread(fname)
        # rescale dynamic range to 0-4095
        try:
            pixelData = ds.pixel_array.astype(np.float32)
        except:
            ds.decompress()
            pixelData = ds.pixel_array.astype(np.float32)

        ds.PixelData = ""
        self.dicomHeaderList.append(ds)

        self.resolution, self.resolution_valid = self.getDicomResolution(ds)
        return pixelData

    # append one image to the internal list
    def appendImage(self, im):
        if type(im) is str:
            try:
                im = self.loadDicomFile(im)
            except:
                print("Error loading file:", im)
                return
        self.imList.append(im)

    def loadNumpyArray(self, data):
        if np.max(data.flat) <= 10: data *= 100
        self.affine = None
        self.resolution_valid = False
        self.medical_volume = None
        self.resolution = [1, 1, 1]

        if len(data.shape) == 2:
            self.appendImage(data)
        elif len(data.shape) == 3:
            for sl in range(data.shape[2]):
                self.appendImage(data[:, :, sl])
        self.medical_volume = voxel.MedicalVolume(data, np.eye(4))

    def load_dosma_volume(self, medical_volume):
        if np.max(medical_volume.volume) < 10:
            medical_volume.volume *= 100
        while np.max(medical_volume.volume) > 10000:
            medical_volume.volume /= 10
        self.medical_volume = medical_volume
        self.resolution = np.array(self.medical_volume.pixel_spacing)
        self.resolution_valid = True
        self.affine = self.medical_volume.affine
        self.imList = ImListProxy(self.medical_volume)
        if medical_volume.headers() is not None:
            header_obj = medical_volume.headers().squeeze()
            if header_obj.shape == ():
                self.dicomHeaderList = [header_obj.item()]
            else:
                self.dicomHeaderList = list(header_obj)
        else:
            self.dicomHeaderList = None

    # load a whole directory of dicom files
    def loadDirectory(self, path):

        medical_volume, affine_valid, title, basepath, basename = dosma_volume_from_path(path, self.fig.canvas)

        self.imList = []
        self.dicomHeaderList = None
        self.medical_volume = None
        self.affine = None
        self.resolution_valid = False
        self.resolution = [1, 1, 1]

        self.load_dosma_volume(medical_volume)
        self.resolution_valid = affine_valid
        self.basename = basename
        self.basepath = basepath
        self.fig.canvas.manager.set_window_title(title)

        if len(self.imList) > 0:
            try:
                self.imPlot.remove()
            except:
                pass
            self.imPlot = None
            self.curImage = 0
            self.displayImage(int(0))
            self.axes.set_xlim(-0.5, self.image.shape[1] - 0.5)
            self.axes.set_ylim(self.image.shape[0] - 0.5, -0.5)



# when called as a script, load all the images in the directory
def main(cls=ImageShow):
    import argparse
    from qtpy.QtWidgets import QApplication, QFileDialog
    parser = argparse.ArgumentParser(description='DICOM Viewer')
    parser.add_argument('-d', '--debounce', type=int, default=int(MOUSE_SCROLL_DEBOUNCE_TIME*1000), help='Mouse scroll debounce time in milliseconds')
    parser.add_argument('path', nargs='?', default=None, help='Path to the directory containing DICOM files')
    args = parser.parse_args()

    if args.path is None:
        app = QApplication(sys.argv)
        options = QFileDialog.Options()
        options |= QFileDialog.ReadOnly
        file_filter = "DICOM Files (*.dcm *.ima);;NIfTI Files (*.nii *.nii.gz);;NumPy Files (*.npz);;All Files (*)"
        file_path, _ = QFileDialog.getOpenFileName(None, "Select File to open", "", file_filter, options=options)
        if file_path:
            args.path = file_path
        else:
            print("No file selected. Exiting.")
            sys.exit(1)

    imFig = cls()
    imFig.scroll_debounce_time = float(args.debounce) / 1000
    imFig.loadDirectory(args.path)
    plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2021 Dafne-Imaging Team
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

import pydicom as dicom
import pydicom.uid as UID

import os
import os.path
import copy
from progress.bar import Bar

def load3dDicom(path):
    allowed_ext = ['.dcm', '.ima']
    basepath = os.path.abspath(path)
    threeDData = None
    info = []
    l = os.listdir(basepath)
    bar = Bar('Loading Dicom', max=len(l))
    threeDlist = []
    for f in sorted(l):
        if os.path.basename(f).startswith('.'): continue
        fname, ext = os.path.splitext(f)
        if ext.lower() in allowed_ext:
            try:
                dFile = dicom.dcmread(os.path.join(basepath, f))
                pixelIma = dFile.pixel_array
                threeDlist.append(np.copy(pixelIma.astype('float32')))
                dFile.PixelData = ""
                info.append(dFile)
            except:
                pass
        bar.next()
    bar.finish()
    if len(threeDlist) == 0:
        return (None, None)
    threeDData = np.stack(threeDlist, axis = 2)
            
    return (threeDData, info)


def save3dDicom(volume, info, path, newSeriesNumber = None, newSeriesDescription = None, newImageComment = None, startImageNumber = 1):
    try:
        os.mkdir(path)
    except:
        pass
    if newSeriesNumber is None:
        try:
            newSeriesNumber = info[0].SeriesNumber
        except:
            newSeriesNumber = 1
        try:
            newSeriesUID = info[0].SeriesInstanceUID
        except:
            newSeriesUID = UID.generate_uid()
    else:
        newSeriesUID = UID.generate_uid()
    
    if newSeriesDescription is None:
        newSeriesDescription = info[0].SeriesDescription
    
    if len(info) == 1:
        dataArray = volume[:,:,0]
        dicomFileData = copy.deepcopy(info[0]) # create a copy of object
        if dataArray.dtype == 'uint16': # check correct format
            dicomFileData.PixelData = dataArray.tostring()
        else:
            dicomFileData.PixelData = dataArray.round().astype('uint16').tostring()
        dicomFileData.SeriesNumber = newSeriesNumber
        dicomFileData.SeriesInstanceUID = newSeriesUID
        dicomFileData.SOPInstanceUID = UID.generate_uid()
        if newImageComment is not None:
            dicomFileData.ImageComment = newImageComment
        
        fName = os.path.join(path, "image0001.dcm") # if one wants to save a part of a dataset
        dicom.dcmwrite(fName, dicomFileData)
    else:
        bar = Bar('Saving Dicom', max=len(info))
        for sl in range(len(info)):
            dataArray = volume[...,sl]
            dicomFileData = copy.deepcopy(info[sl]) # create a copy of object
            if dataArray.dtype == 'uint16': # check correct format
                dicomFileData.PixelData = dataArray.tostring()
            else:
                dicomFileData.PixelData = dataArray.round().astype('uint16').tostring()
            dicomFileData.SeriesNumber = newSeriesNumber
            dicomFileData.SeriesInstanceUID = newSeriesUID
            dicomFileData.SOPInstanceUID = UID.generate_uid()
            if newImageComment is not None:
                dicomFileData.ImageComment = newImageComment
            
            fName = os.path.join(path, "image%04d.dcm" % (sl+startImageNumber)) # if one wants to save a part of a dataset
            dicom.dcmwrite(fName, dicomFileData)
            bar.next()
        bar.finish()


# test
if __name__ == '__main__':
    data, info = load3dDicom('')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2021 Dafne-Imaging Team
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import flexidep
from . import resources
flexidep.standard_install_from_resource(resources, 'runtime_dependencies.cfg')

from .dicom3D import load3dDicom, save3dDicom
from .misc import loadDicomFile, dosma_volume_from_path, realign_medical_volume

def medical_volume_from_path(*args, **kwargs):
    volume, *_ = dosma_volume_from_path(*args, **kwargs)
    return volume

from .padorcut import padorcut


# -*- coding: utf-8 -*-

#  Copyright (c) 2021 Dafne-Imaging Team
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pydicom
import numpy as np
import copy


def flatten_data(d, new_dataset=None):
    if new_dataset is None:
        new_dataset = pydicom.Dataset()
        new_dataset.is_little_endian = d.is_little_endian
        new_dataset.is_implicit_VR = d.is_implicit_VR
        new_dataset.file_meta = copy.deepcopy(d.file_meta)
        try:
            new_dataset.file_meta.MediaStorageSOPClassUID = '1.2.840.10008.5.1.4.1.1.4'
        except:
            print("Cannot set new file meta")
        new_dataset.ensure_file_meta()
    for element in d.iterall():
        if not isinstance(element.value, pydicom.sequence.Sequence):
            new_dataset.add(element)
        # else:
        #    print('Skipping', element)
    return new_dataset


def convert_to_slices(data_in):
    d = copy.copy(data_in)
    slice_data = d[(0x5200, 0x9230)]
    d.pop((0x5200, 0x9230))

    try:
        d.decompress()
    except:
        pass

    pixel_data = d.pixel_array.astype(np.float32)

    if pixel_data.ndim > 2:
        pixel_data = pixel_data.transpose([1, 2, 0])


    d.pop((0x7fe0, 0x0010))
    header_list = []
    for slice_header in slice_data:
        new_slice_header = flatten_data(d)
        flatten_data(slice_header, new_slice_header)
        new_slice_header.NumberOfFrames = 1
        header_list.append(new_slice_header)
    return pixel_data, header_list


def divide_slice_types(pixel_data, header_list):
    def get_frame_type_philips(h):
        return h[(0x2005, 0x1011)].value

    def get_frame_type_generic(h):
        return ';'.join(h.FrameType)

    get_frame_type = get_frame_type_philips
    try:
        current_frame_type = get_frame_type(header_list[0])
    except KeyError:
        get_frame_type = get_frame_type_generic
        current_frame_type = get_frame_type(header_list[0])

    frames_out = {}
    current_frame_number = 0
    last_frame_number = 0
    for current_frame_number, dicom_info in enumerate(header_list):
        new_frame_type = get_frame_type(dicom_info)
        if new_frame_type != current_frame_type:
            new_header = header_list[last_frame_number:current_frame_number]
            new_data = pixel_data[:, :, last_frame_number:current_frame_number]
            frames_out[current_frame_type] = (new_data, new_header)
            current_frame_type = new_frame_type
            last_frame_number = current_frame_number
    # add the last frame type
    new_header = header_list[last_frame_number:current_frame_number+1]
    new_data = pixel_data[:, :, last_frame_number:current_frame_number+1]
    frames_out[current_frame_type] = (new_data, new_header)
    return frames_out


def load_dcm_if_necessary(dicom_file):
    if type(dicom_file) is str:
        d = pydicom.dcmread(dicom_file)
    else:
        d = dicom_file
    return d


def is_enhanced_dicom(dicom_file):
    d = load_dcm_if_necessary(dicom_file)
    try:
        return d.file_meta.MediaStorageSOPClassUID == '1.2.840.10008.5.1.4.1.1.4.1'
    except:
        return False


def is_multi_dicom(dicom_file):
    d = load_dcm_if_necessary(dicom_file)

    try:
        number_of_frames = int(d.NumberOfFrames)
    except:
        number_of_frames = 1

    return number_of_frames > 1


def load_multi_dicom(dicom_file):
    d = load_dcm_if_necessary(dicom_file)

    try:
        number_of_frames = int(d.NumberOfFrames)
    except:
        number_of_frames = 1

    if number_of_frames <= 1:
        return None  # not a multi dicom

    pixel_data, header_list = convert_to_slices(d)
    return divide_slice_types(pixel_data, header_list)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 19 12:48:00 2015

@author: francesco
"""

from .pyDicomView import ImageShow
from .pyDicomView import main as parent_main
from matplotlib.patches import Circle
from matplotlib import pyplot as plt
import math
import numpy as np
try:
    import pyperclip
except:
    print("Pyperclip is not available. Results will not be copied")

class ROIStruct:
    def __init__(self, circle, text, center, radius):
        self.circle = circle
        self.text = text
        self.center = center
        self.radius = radius
        
    def setColor(self, color):
        self.circle.set_edgecolor(color)
        self.text.set_color(color)

class CircleROIClass(ImageShow):
    def __init__(self,  *args, **kwargs):
        ImageShow.__init__(self, *args, **kwargs)
        self.rois = []
        self.colors = ['blue', 'red', 'green', 'yellow', 'magenta', 'cyan', 'indigo']
        self.currentROI = 0
        self.fig.canvas.mpl_connect('close_event', self.print_results)
        self.instructions = 'Shift+click: Draw ROI. Ctrl+click: Delete ROI'
        tb = self.fig.canvas.toolbar
        tb.addSeparator()
        tb.addAction("Copy Results", lambda: self.print_results(None))
        
    def print_results(self, event):
        outStr = 'Image,\troi,\tmean,\tstd'
        for imIndex in range(0, len(self.imList)):
            for roiIndex in range(0, len(self.rois)):
                roi = self.rois[roiIndex]
                im = self.imList[imIndex]
                mean, std = self.calcMeanStd(roi, im)
                outStr += '\n%d,\t%d,\t%.3g,\t%.3g' % (imIndex, roiIndex, mean, std)
        print(outStr)
        try:
            pyperclip.copy(outStr)
        except:
            pass
                
        
        
    # calculates mean and std of the ROI
    def calcMeanStd(self, roi = None, im = None):                
        if roi is None:
            c = self.circleCenter
            r = self.circleRadius
        else:
            c = roi.center
            r = roi.radius

        if im is None:
            im = self.image
        
        ny, nx = im.shape
        ix, iy = np.meshgrid(np.arange(nx), np.arange(ny))
        distance = np.sqrt((ix - c[0])**2 + (iy - c[1])**2)
        # Mask portions of the data array outside of the circle
        maskedImage = np.ma.masked_where(distance > r, im)
        
        return np.ma.mean(maskedImage), np.ma.std(maskedImage)
        
    # generates nice text to be displayed
    def genText(self, mean, std):
        try:
            return 'Mean: %.3g\n Std: %.3g' % (mean, std)
        except:
            return 'Mean: \nStd:'
        
    def leftPressCB(self, event):
        if event.key == 'shift':
            self.createROI(event)
        if event.key == 'control'  or event.key == 'super':
            self.deleteROI(event)
        
    def roiContains(self, roi, event):
        return ((event.xdata-roi.center[0])**2 + (event.ydata-roi.center[1])**2) < roi.radius**2
    
    def deleteROI(self,event):
        #delete all the ROIs under cursor
        for roi in self.rois[:]:
            if self.roiContains(roi, event):
                #remove the roi
                roi.circle.remove()
                roi.text.remove()
                self.rois.remove(roi)
        #update colors
        self.currentROI = len(self.rois)
        for i in range(0, self.currentROI):
            self.rois[i].setColor(self.colors[i])
            self.rois[i].circle.set_label('Roi %d' % i)
        self.axes.legend()
        plt.draw()
            
        
    def createROI(self,event):
        # process only if the event is in the axes
        if not self.imPlot.contains(event):
            return
        
        # check if more rois are allowed
        if self.currentROI >= len(self.colors):
            return    
        
        self.circleCenter = (event.xdata, event.ydata) #self.axes.transData.inverted().transform([event.x, event.y])
        #self.circleCenter = (self.circleCenter[0], self.circleCenter[1]) # convert to tuple
        self.circleRadius = 1
        
        
        self.circle = Circle(self.circleCenter, self.circleRadius, facecolor = 'none', edgecolor = self.colors[self.currentROI], linewidth = 2.0)
        self.axes.add_patch(self.circle)
        mean, std = self.calcMeanStd()
        self.text = self.axes.text(self.circleCenter[0], self.circleCenter[1], self.genText(mean, std), color = self.colors[self.currentROI], verticalalignment = 'top' )
        plt.draw() # draw the circle
        
    def leftReleaseCB(self, event):
        if event.key == 'control'  or event.key == 'super':
            return
        # check if more rois are allowed
        if self.currentROI >= len(self.colors):
            return
        # add the new ROI to the list
        roiStruct = ROIStruct(self.circle, self.text, self.circleCenter, self.circleRadius)
        roiStruct.circle.set_label('Roi %d' % self.currentROI)
        self.rois.append(roiStruct)
        self.currentROI = self.currentROI+1
        self.axes.legend()
        plt.draw()
        
    def leftMoveCB(self,event):
        if event.key != 'shift':
            return
        # check if more rois are allowed
        if self.currentROI >= len(self.colors):
            return
        currentXY = (event.xdata, event.ydata) #self.axes.transData.inverted().transform([event.x, event.y])       
        self.circleRadius = math.sqrt((currentXY[0] - self.circleCenter[0])**2 + (currentXY[1] - self.circleCenter[1])**2)
        self.circle.set_radius(self.circleRadius)
        
        mean, std = self.calcMeanStd()
        self.text.set_position(currentXY)
        self.text.set_text(self.genText(mean, std))
        
        plt.draw()
        

    def refreshCB(self):
        for roi in self.rois:
            mean, std = self.calcMeanStd(roi)
            roi.text.set_text(self.genText(mean, std))
      

# when called as a script, load all the images in the directory
def main():
    parent_main(CircleROIClass)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2021 Dafne-Imaging Team
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import traceback

import pydicom
import pydicom as dicom
import numpy as np
from qtpy.QtWidgets import QInputDialog
from voxel import NiftiReader, DicomReader, MedicalVolume
from voxel.orientation import to_RAS_affine
from scipy.ndimage import map_coordinates
import os

from .multiframe import is_enhanced_dicom, is_multi_dicom, convert_to_slices, load_multi_dicom


class ConversionError(Exception):
    pass

def loadDicomFile(fname):
    ds = dicom.dcmread(fname)
    # rescale dynamic range to 0-4095
    try:
        pixelData = ds.pixel_array.astype(np.float32)
    except:
        ds.decompress()
        pixelData = ds.pixel_array.astype(np.float32)
    ds.PixelData = ""
    return pixelData, ds

def create_affine(sorted_dicoms):
    """
    Function to generate the affine matrix for a dicom series
    This method was based on (http://nipy.org/nibabel/dicom/dicom_orientation.html)

    original (c) icometrix (https://github.com/icometrix/dicom2nifti) - MIT license

    :param sorted_dicoms: list with sorted dicom datasets (from pydicom)
    """

    # Create affine matrix (http://nipy.sourceforge.net/nibabel/dicom/dicom_orientation.html#dicom-slice-affine)
    image_orient1 = np.array(sorted_dicoms[0].ImageOrientationPatient)[0:3]
    image_orient2 = np.array(sorted_dicoms[0].ImageOrientationPatient)[3:6]

    delta_r = float(sorted_dicoms[0].PixelSpacing[0])
    delta_c = float(sorted_dicoms[0].PixelSpacing[1])

    image_pos = np.array(sorted_dicoms[0].ImagePositionPatient)

    last_image_pos = np.array(sorted_dicoms[-1].ImagePositionPatient)

    if len(sorted_dicoms) == 1:
        # Single slice
        step = [0, 0, -1]
    else:
        step = (image_pos - last_image_pos) / (1 - len(sorted_dicoms))

    # check if this is actually a volume and not all slices on the same location
    if np.linalg.norm(step) == 0.0:
        raise ConversionError("NOT_A_VOLUME")

    affine = np.array(
        [[-image_orient1[0] * delta_c, -image_orient2[0] * delta_r, -step[0], -image_pos[0]],
         [-image_orient1[1] * delta_c, -image_orient2[1] * delta_r, -step[1], -image_pos[1]],
         [image_orient1[2] * delta_c, image_orient2[2] * delta_r, step[2], image_pos[2]],
         [0, 0, 0, 1]]
    )
    return affine


def get_nifti_orientation(descriptor: str):
    if descriptor.lower() == 'axial':
        nifti_orient = ('AP', 'RL', 'IS')
    elif descriptor.lower() == 'sagittal':
        nifti_orient = ('SI', 'PA', 'RL')
    elif descriptor.lower() == 'coronal':
        nifti_orient = ('SI', 'RL', 'AP')
    else:
        raise ValueError(f'Unknown orientation descriptor: {descriptor}')

    return nifti_orient


def reorient_data_ui(medical_volume, parent_qobject = None, inplace = False):
    desired_orientation, accept = QInputDialog.getItem(parent_qobject,
                                                       "Nifti loader",
                                                       "Select orientation",
                                                       ['Axial', 'Sagittal', 'Coronal'],
                                                       editable=False)
    if not accept: return None

    mv_out = medical_volume.reformat(get_nifti_orientation(desired_orientation), inplace=inplace)

    return mv_out


def dosma_volume_from_path(path, parent_qobject = None, reorient_data = True, sort='InstanceNumber'):
    medical_volume = None
    affine_valid = False
    basepath = ''
    title = ''

    dicom_ext = ['.dcm', '.ima']
    nii_ext = ['.nii', '.gz']
    npy_ext = ['.npy']
    path = os.path.abspath(path)
    _, ext = os.path.splitext(path)

    basename = os.path.basename(path)

    if ext.lower() in npy_ext:
        data = np.load(path).astype(np.float32)
        medical_volume = MedicalVolume(data, np.eye(4))
        affine_valid = False
        basepath = os.path.dirname(path)
        title = basepath

    elif ext.lower() in nii_ext:
        niiReader = NiftiReader()
        medical_volume = niiReader.load(path)
        if reorient_data:
            medical_volume = reorient_data_ui(medical_volume, parent_qobject, inplace=False)
            if medical_volume is None:
                return

        affine_valid = True
        title = os.path.basename(path)
        basepath = os.path.dirname(path)

    else:  # assume it's dicom
        load_dicom_dir = False
        if os.path.isfile(path):
            basepath = os.path.dirname(path)
            dataset = pydicom.dcmread(path, force=True)
            if is_enhanced_dicom(dataset):
                if is_multi_dicom(dataset):
                    multi_dicom_dataset = load_multi_dicom(dataset)
                    # this is a multi dicom dataset
                    # let the user choose which dataset to load
                    if len(multi_dicom_dataset) == 1:
                        dataset_key = list(multi_dicom_dataset.keys())[0]
                    else:
                        dataset_key, accept = QInputDialog.getItem(parent_qobject,
                                                               "Multi dicom",
                                                               "Choose dataset to load",
                                                               list(multi_dicom_dataset.keys()),
                                                               editable=False)
                        if not accept: return
                    header_list = multi_dicom_dataset[dataset_key][1]
                    data = multi_dicom_dataset[dataset_key][0].astype(np.float32)
                else:
                    # enhanced dicom but not with multiple contrasts
                    data, header_list = convert_to_slices(dataset)

                affine = to_RAS_affine(header_list)
                medical_volume = MedicalVolume(data, affine, header_list)
                affine_valid = True
                title = os.path.basename(path)
                load_dicom_dir = False
            else:
                title = os.path.basename(basepath)
                basename = ''
                load_dicom_dir = True

        elif os.path.isdir(path):
            basename = ''
            basepath = path
            title = basepath
            load_dicom_dir = True

        if load_dicom_dir:
            try:
                if sort == 'None':
                    sort = None
                dr = DicomReader(num_workers=0, group_by=None, sort_by=sort, ignore_ext=True)
                medical_volume = dr.load(basepath)[0]
                affine_valid = True
            except:
                # Error reading with DOSMA. use standard dicom
                print('Error using DOSMA for load')
                #traceback.print_exc()
                l = os.listdir(basepath)
                threeDlist = []
                header_list = []
                for f in sorted(l):
                    if os.path.basename(f).startswith('.'): continue
                    try:
                        dFile = dicom.dcmread(os.path.join(basepath, f), force=True)
                    except:
                        print("Error loading", f)
                        continue
                    dFile.ensure_file_meta()
                    if 'TransferSyntaxUID' not in dFile.file_meta:
                        dFile.file_meta.TransferSyntaxUID = '1.2.840.10008.1.2.1'

                    try:
                        try:
                            dFile.decompress()
                        except ValueError:
                            dFile.NumberOfFrames = 1
                            dFile.decompress()
                    except:
                        print('Error decompressing file')


                    try:
                        pixelIma = dFile.pixel_array
                    except:
                        print('Error loading pixel array for file', f)
                        continue

                    threeDlist.append(np.copy(pixelIma.astype('float32')))
                    dFile.PixelData = ""
                    header_list.append(dFile)
                data = np.stack(threeDlist, axis=2)
                affine = to_RAS_affine(header_list)
                medical_volume = MedicalVolume(data, affine, header_list)
                affine_valid = True

    # more informative title
    if len(path) > 60:
        title = '...' + path[-60:]
    else:
        title = path

    return medical_volume, affine_valid, title, basepath, basename


def realign_medical_volume(source: MedicalVolume, destination: MedicalVolume, interpolation_order: int = 3):
    """Realign this volume to the image space of another volume. Similar to ``reformat_as``,
    except that it supports fine rotations, translations and shape changes, so that the affine
    matrix and extent of the modified volume is identical to the one of the target.


    Args:
        source (MedicalVolume): The volume to realign
        destination (MedicalVolume): The realigned volume will have the same extent and affine matrix of ``destination``.
        interpolation_order (int, optional): spline interpolation order.

    Returns:
        MedicalVolume: The realigned volume.
    """

    target = destination.volume

    # disregard our own slice thickness for the moment
    # calculate the difference from the center of each 3D "partition" and the real center of the 2D slice
    z_offset = 0

    # The following would be need to be used if the origin of the affine matrix was calculated on the middle of the
    # slice with thickness == slice spacing. But centering the converted dataset on the real center of the slice
    # seems to be the norm, hence the z_offset is 0
    # z_offset = (other_thickness/other.pixel_spacing[2]/2 - 1/2) #(other_thickness - other.pixel_spacing[2])/2

    shape_as_range = (np.arange(target.shape[0]),
                      np.arange(target.shape[1]),
                      np.arange(target.shape[2]) + z_offset)

    coords = np.array(np.meshgrid(*shape_as_range, indexing='ij')).astype(float)

    # Add additional dimension with 1 to each coordinate to make work with 4x4 affine matrix
    addon = np.ones([1, coords.shape[1], coords.shape[2], coords.shape[3]])
    coords = np.concatenate([coords, addon], axis=0)  # shape: [4, x, y, z]
    coords = coords.reshape([4, -1])  # shape: [4, x*y*z]

    # build affine which maps from target grid to source grid
    aff_transf = np.linalg.inv(source.affine) @ destination.affine

    # transform the coords from target grid to the space of source image
    coords_src = aff_transf @ coords  # shape: [4, x*y*z]

    # reshape to original spatial dimensions
    coords_src = coords_src.reshape((4,) + target.shape)[:3, ...]  # shape: [3, x, y, z]

    # Will create a image with the spatial size of coords_src (with is target.shape). Each
    # coordinate contains a place in the source image from which the intensity is taken
    # and filled into the new image. If the coordinate is not within the range of the source image then
    # will be filled with 0.
    src_transf_data = map_coordinates(source.volume, coords_src, order=interpolation_order)

    mv = MedicalVolume(src_transf_data, destination.affine, destination.headers())

    return mv

